@@ -1,21 +1,3 @@
-class RequestHeaders(object):
-    def __init__(self, headers: dict):
-        self.this = headers
-
-    def __getattr__(self, key: str):
-        """convert snake_cased headers to dashed and retrieve from headers dict
-        """
-        key = key.encode() if isinstance(key, str) else key
-
-        #ignore for custom x-headers NOTE to put this in documentation later...
-        k = key if key.lower().startswith(b'x-') else key.replace(b'_', b'-')
-        return self.this.get(k)
-
-    def get(self, key):
-        key = key.encode() if isinstance(key, str) else key
-        return self.this.get(key)
-
-
 class HttpRequest():
     def __init__(self, scope, body, receive, metadata=None):
         self._scope = scope
@@ -39,7 +21,7 @@ class HttpRequest():
                 if isinstance(current_value, list): current_value.append(value)
                 else: qsd[key] = [current_value, value]
         return qsd
-    
+
     @property
     def body(self):
         return self._body.get('body')
@@ -51,6 +33,10 @@ class HttpRequest():
             for header in self._scope.get('headers'):
                 self._headers[header[0]] = header[1]
         return self._headers
+     
+    @property
+    def method(self):
+        return self._scope.get('method')
 
     @property
     def params(self):
@@ -66,7 +52,7 @@ class HttpRequest():
 
     @property
     def querystring(self):
-        return self._scope.get('query_string', {})
+        return self._scope.get('query_string', '')
     
     @property
     def subdomain(self):
@@ -77,4 +63,5 @@ class HttpRequest():
         return self._scope.get('path')
 
     async def stream(self):
+        """TODO: Revisit and implement with appropriate testing"""
         return await self._receive()
