@@ -2,17 +2,19 @@ import sys
 
 from collections import deque
 from inspect import iscoroutinefunction
-from ipaddress import ip_address
 
 from uvicorn import run
 from typing import Callable, Generic, TypeVar
 
 from .constants import (
     CONNECT,
+    DEFAULT,
     DELETE,
     GET,
     HEAD,
+    INITIALIZATION_MESSAGE,
     MESSAGE_NOT_FOUND,
+    METHODS,
     METHOD_CONNECT,
     METHOD_DELETE,
     METHOD_GET,
@@ -30,7 +32,8 @@ from .constants import (
     URL_ERROR_MESSAGE
 )
 
-from .errors import AbortException, UrlError
+from .errors import AbortException, SubdomainError, UrlDuplicateError, UrlError
+from .utils import preprocessor
 from .request import HttpRequest
 from .response import ResponseWriter
 from .context import Context
@@ -41,19 +44,25 @@ methods = ['get', 'post', 'put', 'delete', 'connect', 'head', 'options', 'patch'
 
 Handler = Callable[[HttpRequest, ResponseWriter], object]
 
-
-DEFAULT = 'www'
 MARK_PRESENT = 'yes father... (RIP Rev. Angus Fraser...)'
 SEPARATOR = INDEX = "/"
 
 
-def isparamx(r: str):
+def _get_configuration(configurator=None):
+    if not configurator: return {}
+    if isinstance(configurator, dict): return configurator
+    return configurator()
+
+
+def _isparamx(r: str):
     return (':', r[1:],) if r.startswith(':') else (r, None,)
 
 
-def isvariable(r: str):
-    newr = ':' if r.startswith(':') else r
-    return newr, newr != r or r == '*'
+def _notify(width=80, event='startup'): #pragma: nocover
+    drawline = lambda: print('=' * width)
+    drawline()
+    print(f'NOTE: The `LAST` {event} func above failed and prevented others from running')
+    drawline()
 
 
 class Route(object):
@@ -63,21 +72,6 @@ class Route(object):
         self.handler = handler
         self.children = {}
 
-    def add(self, pathstring: deque, handler):
-        # We can do this with confidence because we use cache to ensure url mask not used twice
-        node = self
-        while pathstring:
-            route = pathstring.popleft()
-            url, parameterized = isparamx(route)
-            current_node = node.children.get(route)
-            if not current_node:
-                url, func = None, None
-                current_node = Route(url, func)
-                node.children[route] = current_node
-            node = current_node
-            node.parameterized = parameterized
-        node.handler = handler
-
     def match(self, routes: deque, r: HttpRequest):
         matched: str = None
         route_at_deviation = None
@@ -98,7 +92,7 @@ class Route(object):
                     continue
 
                 wildcard = node.children.get('*')
-                if wildcard:
+                if wildcard:  #pragma: nocover
                     r.params = '*', '/'.join([route, *routes])
                     return wildcard.route, wildcard.handler
 
@@ -118,9 +112,6 @@ class Route(object):
     def not_found(self, r: HttpRequest, w: ResponseWriter, c: Context):
         w.status = 404
         w.body = b'Not found'
-    
-    def retrieve(self, match: str):
-        pass
 
 
 class Routes(object):
@@ -132,8 +123,15 @@ class Routes(object):
         self.routes = {}
 
     def add(self, method: str, route: str, handler: Callable):
+        """
+        method: one of POST, GET, OPTIONS... etc - i.e. the HTTP method
+        route: the route url/endpoint
+        handler: function corresponding to the signature of a routerling handler
+        """
         # ensure the method and route combo has not been already registered
-        assert self.cache.get(method, {}).get(route) is None
+        try: assert self.cache.get(method, {}).get(route) is None
+        except AssertionError: raise UrlDuplicateError
+
         self.cache[method][route] = MARK_PRESENT
 
         # here we check and set the root to be a route node i.e. / with no handler
@@ -155,17 +153,17 @@ class Routes(object):
         stop_at = len(routes) - 1
 
         for index, routerling in enumerate(routes):
-            routerling, parameterized = isparamx(routerling)
-            new_route_node = route_node.children.get(routerling)
+            _routerling, _parameterized = _isparamx(routerling)
+            new_route_node = route_node.children.get(_routerling)
             if not new_route_node:
                 new_route_node = Route(None, None)
-                route_node.children[routerling] = new_route_node
+                route_node.children[_routerling] = new_route_node
             
             route_node = new_route_node
-            route_node.parameterized = parameterized
+            route_node.parameterized = _parameterized
 
             if index == stop_at:
-                assert route_node.handler is None, f'Handler already registered for route: ${routerling}'
+                assert route_node.handler is None, f'Handler already registered for route: ${_routerling}'
                 route_node.route = route
                 route_node.handler = handler
     
@@ -199,15 +197,15 @@ class Routes(object):
         for route in routes:...
         return None, None
 
-    async def handle(self, scope, receive, send, metadata=None):
+    async def handle(self, scope, receive, send, metadata=None, application=None):
         """
         Traverse internal route tree and use appropriate method
         """
         body = await receive()
 
         r = HttpRequest(scope, body, receive, metadata)
-        w = ResponseWriter(scope)
-        c = Context()
+        w = ResponseWriter()
+        c = Context(application)
 
         method = scope.get('method')
         matched = None
@@ -224,7 +222,7 @@ class Routes(object):
             return w
 
         route = scope.get('path')
-        if route == SEPARATOR:
+        if route == SEPARATOR: #pragma: nocover
             matched = route_node.route # same as = SEPARATOR
             handler = route_node.handler
         else:
@@ -241,6 +239,9 @@ class Routes(object):
 
             # call request handler
             if w._abort: raise AbortException
+            try: handler.__requesthandler__
+            except: pass
+            else: handler = handler.__call__
             if iscoroutinefunction(handler): await handler(r, w, c)
             else: handler(r, w, c)
 
@@ -253,6 +254,24 @@ class Routes(object):
         # i.e. hook lookup is constant but running many hooks will
         # increase the time it takes before the response is released to the network
         return w
+    
+    def remove(self, method: str, route: str):
+        assert method in METHODS
+        route_node = self.routes.get(method)
+        if not route_node: return
+        if not route_node.children: return
+
+        routes = route.strip(SEPARATOR).split(SEPARATOR)
+        stop_at = len(routes) - 1
+        for index, routerling in enumerate(routes):
+            _routerling, _parameterized = _isparamx(routerling)
+            route_node = route_node.children.get(_routerling)
+            if not route_node:
+                return
+            if index == stop_at:
+                route_node.route = None
+                route_node.handler = None
+                self.cache[method][route] = None
 
     async def xhooks(self, hookstore, matched, r: HttpRequest, w: ResponseWriter, c: Context):
         # traverse the before tree - changed to tree to match routes tree
@@ -266,18 +285,36 @@ class Routes(object):
             hooks.update(hookstore.get(f'/{joinedparts}{_}*', []))
         for hook in hooks:
             if w._abort: raise AbortException
-            hook(r, w, c)
+            if iscoroutinefunction(hook): await hook(r, w, c)
+            else: hook(r, w, c)
 
 
 class Router(object):
-    def __init__(self):
+    def __init__(self, configurator=None):
+        self.finalized = False
+        self.initializers = set()
+        self.deinitializers = set()
         self.subdomains = {}
         self.subdomains[DEFAULT] = Routes()
+        self._buckets = {}
+        self._configuration = _get_configuration(configurator)
 
     async def __call__(self, scope, receive, send):
-        metadata = self.__subdomain__(scope)
-        engine = self.subdomains.get(metadata[0])
-        response = await engine.handle(scope, receive, send, metadata)
+        if scope['type'] == 'lifespan':
+            while True:
+                message = await receive()
+                if message['type'] == 'lifespan.startup':
+                    try: await self._register()
+                    except: _notify()
+                    await send({'type': 'lifespan.startup.complete'})
+                elif message['type'] == 'lifespan.shutdown':
+                    try: await self._unregister()
+                    except: _notify(event='shutdown')
+                    await send({'type': 'lifespan.shutdown.complete'})
+
+        metadata = preprocessor(scope)
+        engine = self.subdomains.get(metadata[0]) or self.subdomains.get('*')
+        response = await engine.handle(scope, receive, send, metadata, self)
         await send({
             'type': 'http.response.start',
             'headers': response.headers,
@@ -287,48 +324,34 @@ class Router(object):
             'type': 'http.response.body',
             'body': response.body
         })
-    
-    def __subdomain__(self, scope):
-        headers = {}
-
-        for header in scope.get('headers'):
-            key, value = header
-            headers[key.decode()] = value.decode()
-            
-        host: bytes = headers.get('host')
-        if host.startswith('http://'): host = host.replace('http://', '')
-        else: host = host.replace('https://', '')
-        host = host.rsplit(':')[0]
-        try: ip_address(host)
-        except: return DEFAULT, headers
-        parts = host.split('.', 2)
-        has_subdomain = len(parts) > 2
-        return (parts[0], headers,) if has_subdomain else (DEFAULT, headers,)
 
     def abettor(self, method: str, route: str, handler: Handler, subdomain=DEFAULT):
         if not route.startswith('/'): raise UrlError
         engine = self.subdomains.get(subdomain)
         if not isinstance(engine, Routes):
-            sys.exit('Error: subdomain not registered on router')
+            raise SubdomainError
         engine.add(method, route, handler)
 
     def AFTER(self, route: str, handler: Callable, subdomain=DEFAULT):
         if not route.startswith('/'): raise UrlError(URL_ERROR_MESSAGE)
         engine = self.subdomains.get(subdomain)
-        if not isinstance(engine, Routes):
+        if not isinstance(engine, Routes): #pragma: nocover
             raise NameError('Subdomain does not exist - register subdomain on router first')
         engine.after = route, handler
 
     def BEFORE(self, route: str, handler: Callable, subdomain=DEFAULT):
         if not route.startswith('/'): raise UrlError(URL_ERROR_MESSAGE)
         engine = self.subdomains.get(subdomain)
-        if not isinstance(engine, Routes):
+        if not isinstance(engine, Routes): #pragma: nocover
             raise NameError('Subdomain does not exist - register subdomain on router first')
         engine.before = route, handler
 
     def CONNECT(self, route: str, handler: Callable[[HttpRequest, ResponseWriter], object], subdomain=DEFAULT):
         self.abettor(METHOD_CONNECT, route, handler, subdomain)
 
+    def CONFIG(self, config):
+        return self._configuration[config]
+
     def DELETE(self, route: str, handler: Callable, subdomain=DEFAULT):
         self.abettor(METHOD_DELETE, route, handler, subdomain)
     
@@ -354,8 +377,63 @@ class Router(object):
     def TRACE(self, route: str, handler: Callable, subdomain=DEFAULT):
         self.abettor(METHOD_TRACE, route, handler, subdomain)
 
-    def listen(self, host='localhost', port='8701', debug=DEFAULT):
-        # implement development server? uvicorn already present as dependency so not in python library
+    def ONCE(self, *args):
+        arguments = len(args)
+        error_message = 'ONCE requires a callable argument as default'
+        help_message = 'If 2 arguments provided: first: str = `startup` or `shutdown` AND second: Callable'
+        try: assert arguments <= 2
+        except: raise TypeError('ONCE function received more than 2 arguments')
+
+        if arguments == 1:
+            first = args[0]
+            try: assert isinstance(first, Callable)
+            except AssertionError: raise TypeError(error_message)
+            self.initializers.add(first)
+        else:
+            first, second = args
+
+            try: assert first in ['startup', 'shutdown']
+            except AssertionError: raise ValueError(help_message)
+
+            try: assert isinstance(second, Callable)
+            except AssertionError: raise TypeError(error_message)
+
+            if first == 'startup': self.initializers.add(second)
+            else: self.deinitializers.add(second)
+
+    async def _register(self):
+        i = len(self.initializers)
+        while self.initializers:
+            initializer, c = self.initializers.pop(), len(self.initializers)
+            index = i - c
+            print(f'({index}): ', initializer.__name__, '\n')
+            if iscoroutinefunction(initializer): await initializer(self)
+            else: initializer(self)
+    
+    async def _unregister(self):
+        i = len(self.deinitializers)
+        while self.deinitializers:
+            deinitializer, c = self.deinitializers.pop(), len(self.deinitializers)
+            index = i - c
+            print(f'({index}): ', deinitializer.__name__, '\n')
+            if iscoroutinefunction(deinitializer): await deinitializer(self)
+            else: deinitializer(self)
+    
+    def keep(self, key, value):
+        self._buckets[key] = value
+
+    def unkeep(self, key):
+        value = self._buckets[key]
+        del self._buckets[key]
+        return value
+
+    def peek(self, key):
+        try: value = self._buckets[key]
+        except KeyError: return None
+        else: return value
+
+    def listen(self, host='localhost', port='8701', debug=DEFAULT): #pragma: nocover
+        # repurpose this for websockets?
         pass
 
     def subdomain(self, subdomain: str):
